/**
 * ForceDirectedGraph
 * by Sulav Malla 
 * 
 * A Class to create a force directed graph
 *
 */

class ForceDirectedGraph {
  
  // All nodes and links in the graphs
  JSONArray allNodes;
  JSONArray allLinks;
  
  JSONObject clickedNode; // for interactions
  
  // HashMap to store all nodes of the graph.
  // Key is id (String) and value is 
  // {id, group, pos_x, pos_y, vel_x, vel_y, mass} (JSONObject)
  // This is for storing and faster lookup 
  // of node position and velocity.
  HashMap<String, JSONObject> nodes;
    
  // Chart embellishments and FDG properties
  float    nodeMass;
  int      nodeSize;
  color    nodeFillColor;
  color    nodeStrokeColor;
  int      linkWidth;
  color    linkStrokeColor;
  int      figWidth, figHeight;
  
  float    timeStep;
  float    repulsionConstant;
  float    attractionConstant;
  float    springRestLength;
  float    dampingCoefficient;
  
  /**
   * Constructor to initialize variables
   */
  ForceDirectedGraph() {
    
    // variable initialization
    allNodes           = null;
    allLinks           = null;
    clickedNode        = null;
    nodes              = new HashMap<String, JSONObject>();
    nodeMass           = 1;
    nodeSize           = 5;
    nodeFillColor      = color(44, 123, 181);
    nodeStrokeColor    = color(0);
    linkWidth          = 1;
    linkStrokeColor    = color(153);
    figWidth           = 200;
    figHeight          = 200;
    timeStep           = 0.5;
    repulsionConstant  = 2.0;
    attractionConstant = 2.0;
    springRestLength   = 15;
    dampingCoefficient = 0.8;
  }
  
  /**
   * Setter funtions to set attributes
   */
  void setNodeMass(float value) {              nodeMass = value;        }
  void setNodeSize(int value) {                nodeSize = value;        }
  void setNodeFillColor(color value) {    nodeFillColor = value;        }
  void setNodeStrokeColor(color value) {nodeStrokeColor = value;        }
  void setLinkWidth(int value) {              linkWidth = value;        }
  void setLinkStrokeColor(color value) {linkStrokeColor = value;        }
  void setTimeStep(float value) {              timeStep = value;        }
  void setRepulsionConstant(float value) {  repulsionConstant = value;        }
  void setAttractionConstant(float value) {attractionConstant = value;        }
  void setSpringRestLength(float value) {    springRestLength = value;        }
  void setDampingCoefficient(float value) {dampingCoefficient = value;        }
  void setFigSize(int _figWidth, int _figHeight) {
    figWidth  = _figWidth;
    figHeight = _figHeight;
  }
  
  
  void setData(JSONObject data) {
    // Get the data as JSON object
    allNodes = data.getJSONArray("nodes");
    allLinks = data.getJSONArray("links");
    
    // Make a HashMap of the nodes to store position
    for (int i = 0; i<allNodes.size(); i++) {
      JSONObject node = allNodes.getJSONObject(i);
      
      // set initial mass, position and velocity of the nodes
      node.setFloat("mass", nodeMass);
      node.setFloat("pos_x", random(300,500));
      node.setFloat("pos_y", random(300,500));
      node.setFloat("vel_x", 0);
      node.setFloat("vel_y", 0);
      String id = node.getString("id");
      
      // Store this id:node pair in the HashMap
      nodes.put(id, node);
    }
  }

  
  /**
   * Funtion to draw the actual force directed graph
   */
  void draw() {
    if ((allNodes == null) || (allLinks == null)) {
      return;
    }
    
    // draw the links
    int x1, y1;
    int x2, y2;
    JSONObject link = null;
    JSONObject sourceNode = null;
    JSONObject targetNode = null;
    stroke(linkStrokeColor);
    strokeWeight(linkWidth);
    // Loop over all links and draw them
    for (int i = 0; i<allLinks.size(); i++) {
      link = allLinks.getJSONObject(i);
      sourceNode = nodes.get(link.getString("source"));
      targetNode = nodes.get(link.getString("target"));
      
      x1 = int(sourceNode.getFloat("pos_x"));
      y1 = int(sourceNode.getFloat("pos_y"));
      
      x2 = int(targetNode.getFloat("pos_x"));
      y2 = int(targetNode.getFloat("pos_y"));
      
      line(x1, y1, x2, y2);
    }

    // draw the nodes
    int x_position;
    int y_position;
    int group;
    JSONObject node;

    ellipseMode(CENTER);
    stroke(nodeStrokeColor);
    strokeWeight(1);
    // fill(nodeFillColor);
    // Loop over and draw each node
    for (int i = 0; i<allNodes.size(); i++) {
      node = allNodes.getJSONObject(i);
      x_position = int(node.getFloat("pos_x"));
      y_position = int(node.getFloat("pos_y"));
      group = int(node.getInt("group"));
      switch(group) {
        case 0:
          fill(color(152, 223, 138));
          break;
        case 1:
          fill(color(31, 119, 180));
          break;
        case 2:
          fill(color(174, 199, 232));
          break;
        case 3:
          fill(color(255, 127, 14));
          break;
        case 4:
          fill(color(255, 187, 120));
          break;
        case 5:
          fill(color(44, 160, 44));
          break;
        case 6:
          fill(color(214, 39, 40));
          break;
        case 7:
          fill(color(255, 152, 150));
          break;
        case 8:
          fill(color(148, 103, 189));
          break;
        case 9:
          fill(color(197, 176, 213));
          break;
        case 10:
          fill(color(140, 86, 75));
          break;
        default:
          fill(nodeFillColor);
      }
      ellipse(x_position, y_position, nodeSize, nodeSize);
    }
    
    // Update position and velocity of nodes
    updateNodes();
    
    drawLegend();
  }
  
  
  /**
   * Funtion to do all the force calculations
   * and update a node's position and velocity
   */
  void updateNodes() {
    PVector delVelocity = new PVector(0.0, 0.0);
    PVector force = new PVector(0.0, 0.0);;
    float   v_x, v_y;  // variables to update velocity
    JSONObject node;
        
    // Iterate over all nodes to calculate changes
    for (int i = 0; i<allNodes.size(); i++) {
      node = allNodes.getJSONObject(i);
      
      if(node == clickedNode)   continue;
      force = PVector.add(calcAttractiveForce(node), calcRepulsiveForce(node));
      //force = calcRepulsiveForce(node);
      delVelocity = PVector.mult(force, (timeStep/node.getFloat("mass")));
      // Update velocity
      v_x = node.getFloat("vel_x");
      v_y = node.getFloat("vel_y");
      v_x = (v_x + delVelocity.x) * dampingCoefficient;
      v_y = (v_y + delVelocity.y) * dampingCoefficient;
      
      node.setFloat("vel_x", v_x);
      node.setFloat("vel_y", v_y);
    }
    //println(force);
    //println(delVelocity.x+","+delVelocity.y+","+delVelocity.z);
    
    // Make another pass over all nodes to update position
    float   p_x, p_y;  // variables to update position
    for (int i = 0; i<allNodes.size(); i++) {
      node = allNodes.getJSONObject(i);
      v_x = node.getFloat("vel_x");
      v_y = node.getFloat("vel_y");
      
      
      
      // Update position
      p_x = node.getFloat("pos_x");
      p_y = node.getFloat("pos_y");
      p_x = p_x + (v_x * timeStep);
      p_y = p_y + (v_y * timeStep);
      // Bound the position so that it is inside the screen
      if(p_x < 0)  p_x = 0;
      if(p_y < 0)  p_y = 0;
      if(p_x > figWidth)   p_x = figWidth;
      if(p_y > figHeight)  p_y = figHeight;
      node.setFloat("pos_x", p_x);
      node.setFloat("pos_y", p_y);
    }
  }
  
  
  /**
   * Funtion to calculate total repulsive force to this node1.
   * A node experiences repulsion from all other nodes.
   * Returns the sum of all repulsive force
   */
  PVector calcRepulsiveForce(JSONObject node1) {
    float m1 = node1.getFloat("mass");
    PVector P = new PVector(node1.getFloat("pos_x"), node1.getFloat("pos_y"));
    
    JSONObject node2;
    float m2;
    PVector Q = new PVector(0.0, 0.0);
    
    PVector totalForce = new PVector(0.0, 0.0);
    float distance;
    float force;
    
    // Iterate over all nodes to calculate repulsion
    for (int i = 0; i<allNodes.size(); i++) {
      node2 = allNodes.getJSONObject(i);
      m2 = node2.getFloat("mass");
      Q.set(node2.getFloat("pos_x"), node2.getFloat("pos_y"));
      
      distance = PVector.dist(P, Q);
      if(distance == 0.0)  continue;
      // to avoid distance from being a very small number
      // causing force to be very large
      distance += 0.01;
      force = (repulsionConstant * m1 * m2) / (distance*distance);
      totalForce.add(PVector.sub(P, Q).normalize().mult(force));
    }
    
    return totalForce;
  }
  
  
  /**
   * Funtion to calculate total attractive force to this node1.
   * A node experiences attraction from all other nodes connected to it.
   * Returns the sum of all attractive force
   */
  PVector calcAttractiveForce(JSONObject node1) {
    PVector totalForce = new PVector(0.0, 0.0);
    float distance;
    float force;
    
    PVector P = new PVector(node1.getFloat("pos_x"), node1.getFloat("pos_y"));
    PVector Q = new PVector(0.0, 0.0);
    
    JSONObject link = null;
    JSONObject node2 = null;
    
    // Iterate over all links to check if this node is a part of it.
    for (int i = 0; i<allLinks.size(); i++) {
      link = allLinks.getJSONObject(i);
      if((node1.getString("id").equals(link.getString("source")) == true) ||
         (node1.getString("id").equals(link.getString("target")) == true)){
        // This link is connected to the node
        if(node1.getString("id").equals(link.getString("source")) == true) {
          node2 = nodes.get(link.getString("target"));
        } else {
          node2 = nodes.get(link.getString("source"));
        }
        
        Q.set(node2.getFloat("pos_x"), node2.getFloat("pos_y"));
        distance = PVector.dist(P, Q);
        if((distance - springRestLength) <= 0.0)  continue;
        force = attractionConstant * (distance - springRestLength);
        totalForce.add(PVector.sub(Q, P).normalize().mult(force));
        
      }
    }
    
    return totalForce;    
  }
  
  
  void mouseWasPressed(int x, int y) {
    JSONObject node;
    int offset = nodeSize / 2;

    // Iterate over all nodes to check if it was clicked
    for (int i = 0; i<allNodes.size(); i++) {
      node = allNodes.getJSONObject(i);
      // Check if mouse is over this node
      if ((x >= node.getFloat("pos_x") - offset) && (x <= node.getFloat("pos_x") + offset)) {
        if ((y >= node.getFloat("pos_y") - offset) && (y <= node.getFloat("pos_y") + offset)) {
          node.setFloat("vel_x", 0);
          node.setFloat("vel_y", 0);
          clickedNode = node;
          break;
        }
      }
    }
  }


  void mouseWasDragged(int x, int y) {
    if(clickedNode != null) {
      clickedNode.setFloat("vel_x", 0);
      clickedNode.setFloat("vel_y", 0);
      
      // Bound the position so that it is inside the screen
      if(x < 0)  x = 0;
      if(y < 0)  y = 0;
      if(x > figWidth)   x = figWidth;
      if(y > figHeight)  y = figHeight;
      
      clickedNode.setFloat("pos_x", x);
      clickedNode.setFloat("pos_y", y);
    }
  }


  void mouseWasReleased() {
    clickedNode = null;
  }
  
  void drawLegend() {
    // draw the legend
    int legend_width  = 83;
    int legend_height = 225;
    int legend_left   = figWidth - (legend_width*2);
    int legend_top    = 10;
    int x_position    = legend_left + 15;
    int y_position    = legend_top + 10;
    int spacing       = 20;
    textFont(font_arial, 12);        
    textAlign(LEFT);
    stroke(0);
    fill(255);
    rect(legend_left , legend_top, legend_width, legend_height);
    
    fill(color(152, 223, 138));
    ellipse(x_position, y_position, nodeSize, nodeSize);
    fill(0);
    text("Group 0", x_position + 15, y_position + 5);
    y_position += spacing;
    
    fill(color(31, 119, 180));
    ellipse(x_position, y_position, nodeSize, nodeSize);
    fill(0);
    text("Group 1", x_position + 15, y_position + 5);
    y_position += spacing;
    
    fill(color(174, 199, 232));
    ellipse(x_position, y_position, nodeSize, nodeSize);
    fill(0);
    text("Group 2", x_position + 15, y_position + 5);
    y_position += spacing;
    
    fill(color(255, 127, 14));
    ellipse(x_position, y_position, nodeSize, nodeSize);
    fill(0);
    text("Group 3", x_position + 15, y_position + 5);
    y_position += spacing;

    fill(color(255, 187, 120));
    ellipse(x_position, y_position, nodeSize, nodeSize);
    fill(0);
    text("Group 4", x_position + 15, y_position + 5);
    y_position += spacing;

    fill(color(44, 160, 44));
    ellipse(x_position, y_position, nodeSize, nodeSize);
    fill(0);
    text("Group 5", x_position + 15, y_position + 5);
    y_position += spacing;

    fill(color(214, 39, 40));
    ellipse(x_position, y_position, nodeSize, nodeSize);
    fill(0);
    text("Group 6", x_position + 15, y_position + 5);
    y_position += spacing;

    fill(color(255, 152, 150));
    ellipse(x_position, y_position, nodeSize, nodeSize);
    fill(0);
    text("Group 7", x_position + 15, y_position + 5);
    y_position += spacing;

    fill(color(148, 103, 189));
    ellipse(x_position, y_position, nodeSize, nodeSize);
    fill(0);
    text("Group 8", x_position + 15, y_position + 5);
    y_position += spacing;

    fill(color(197, 176, 213));
    ellipse(x_position, y_position, nodeSize, nodeSize);
    fill(0);
    text("Group 9", x_position + 15, y_position + 5);
    y_position += spacing;

    fill(color(140, 86, 75));
    ellipse(x_position, y_position, nodeSize, nodeSize);
    fill(0);
    text("Group 10", x_position + 15, y_position + 5);
    y_position += spacing;

  }
}